#
#  nn/utils/__init__.py
#  bxtorch
#
#  Created by Oliver Borchert on May 10, 2019.
#  Copyright (c) 2019 Oliver Borchert. All rights reserved.
#  

from .model import Config, Configurable
from .predictor import Predictor
