#
#  job/__init__.py
#  bxtorch
#
#  Created by Oliver Borchert on May 15, 2019.
#  Copyright (c) 2019 Oliver Borchert. All rights reserved.
#  

from .job import Job
import bxtorch.job.arguments as args
