#
#  nn/data/composite.py
#  bxtorch
#
#  Created by Oliver Borchert on May 10, 2019.
#  Copyright (c) 2019 Oliver Borchert. All rights reserved.
#  

class ZipDataLoader:
    """
    Data loader which yields the elements of multiple data loaders.
    """

    pass
