#
#  nn/training/__init__.py
#  bxtorch
#
#  Created by Oliver Borchert on May 20, 2019.
#  Copyright (c) 2019 Oliver Borchert. All rights reserved.
#

from .parallel import Hogwild
from .supervised import SupervisedTrainer
from .trainer import BaseTrainer
